@@ -35,18 +35,23 @@
  *               -- Roy Keene [300820031505] <tcpcgi@rkeene.org>    
  */
 
+#include <string.h>
+#define _XOPEN_SOURCE_EXTENDED
+#include <netinet/in.h>
 #include <sys/socket.h>
 #include <sys/types.h>
+#include <arpa/inet.h>
 #include <sys/time.h>
 #include <sys/poll.h>
 #include <sys/wait.h>
 #include <unistd.h>
 #include <signal.h>
-#include <string.h>
 #include <stdlib.h>
 #include <stdio.h>
 #include <ctype.h>
 #include <fcntl.h>
+#include <netdb.h>
+#include <errno.h>
 #include <time.h>
 
 #include "tcpcgi.h"
@@ -64,7 +69,6 @@ struct sockinfo {
 	signed int fd;
 };
 
-int alarm_freq=0;
 int use_post=1;
 int local_only=0;
 int poll_time_min=TCPCGI_CLIENT_POLLTIME_MIN;
@@ -115,7 +119,7 @@ char *hexcode(void *data, int datalen) {
 	return(ret);
 }
 
-void *tcpcgi_getdata(char *http_host, int http_port, char *http_path, char *cmd, char *session, char *dest, void *data, int *datalen, signed int *status) {
+void *tcpcgi_getdata(char *http_host, int http_port, char *http_path, char *cmd, char *session, char *dest, void *data, int *datalen, signed int *status, int proto_type) {
 	FILE *fp;
 	char *hex_cmd=NULL, *hex_session=NULL, *hex_data=NULL, *hex_path=NULL, *hex_dest=NULL;
 	char buf[8192], *retbuf=NULL;
@@ -123,8 +127,10 @@ void *tcpcgi_getdata(char *http_host, int http_port, char *http_path, char *cmd,
 	int sink=0;
 	unsigned int datalen_loc;
 
+	PRINTERR("Entering tcpcgi_getdata()");
 	if (http_host==NULL || http_path==NULL || cmd==NULL || session==NULL) {
 		PRINTERR("Invalid parameters passed.");
+		if (status) *status=-1;
 		return(NULL);
 	}
 
@@ -136,21 +142,36 @@ void *tcpcgi_getdata(char *http_host, int http_port, char *http_path, char *cmd,
 	hex_dest=hexcode(dest, strlen(dest?dest:""));
 
 	fd=createconnection_tcp(http_host, http_port);
+	if (fd<0) {
+		PERROR("connect");
+		PRINTERR("Couldn't connect!");
+		if (status) *status=-1;
+		return(NULL);
+	}
 	fp=fdopen(fd, "r+");
-	snprintf(buf, sizeof(buf), "cmd=%s&session=%s&dest=%s&data=%s&ts=%i", hex_cmd, hex_session, hex_dest, hex_data, (unsigned int) time(NULL));
+	if (!fp) {
+		PERROR("fdopen");
+		if (status) *status=-1;
+		return(NULL);
+	}
+	snprintf(buf, sizeof(buf), "cmd=%s&session=%s&dest=%s&data=%s&ts=%i&proto=%s", hex_cmd, hex_session, hex_dest, hex_data, (unsigned int) time(NULL), (proto_type==SOCK_DGRAM)?"udp":"tcp");
 	buf[sizeof(buf)-1]='\0';
 	if (use_post) {
-		fprintf(fp, "POST %s HTTP/1.0\r\nHost: %s\r\nContent-length: %i\r\n\r\n%s", hex_path, http_host, strlen(buf), buf);
-		PRINTERR("POST http://%s:%i%s?[%i]%s HTTP/1.0\n", http_host, http_port, hex_path, strlen(buf), buf);
+		fprintf(fp, "POST %s HTTP/1.0\r\nHost: %s\r\nContent-length: %i\r\n\r\n%s\r\n", hex_path, http_host, (int) strlen(buf), buf);
+		PRINTERR("POST http://%s:%i%s?[%i]%s HTTP/1.0\n", http_host, http_port, hex_path, (int) strlen(buf), buf);
 	} else {
 		fprintf(fp, "GET http://%s:%i%s?%s HTTP/1.0\r\n\r\n", http_host, http_port, hex_path, buf);
-		PRINTERR("GET http://%s:%i%s?[%i]%s HTTP/1.0\n", http_host, http_port, hex_path, strlen(buf), buf);
+		PRINTERR("GET http://%s:%i%s?[%i]%s HTTP/1.0\n", http_host, http_port, hex_path, (int) strlen(buf), buf);
 	}
 	fflush(fp);
 
 	if (status) *status=0;
 	if (datalen) *datalen=0;
+	PRINTERR("Scheduling panic alarm.");
+	alarm(120); /* In case of emergency, please panic. */
 	if (fgets(buf, sizeof(buf), fp)==NULL) return(NULL);
+	PRINTERR("Unscheduling panic alarm.");
+	alarm(0); /* Panic no more. */
 	if (strstr(buf," 40")) {
 		PRINTERR("File not available error, %s", buf);
 		if (hex_cmd) free(hex_cmd);
@@ -170,8 +191,10 @@ void *tcpcgi_getdata(char *http_host, int http_port, char *http_path, char *cmd,
 		if (hex_path) free(hex_path);
 		sleep(1); /* Primitive backoff */
 		fclose(fp);
-		return(tcpcgi_getdata(http_host, http_port, http_path, cmd, session, dest, data, datalen, status));
+		return(tcpcgi_getdata(http_host, http_port, http_path, cmd, session, dest, data, datalen, status, SOCK_STREAM));
 	}
+	PRINTERR("Scheduling panic alarm.");
+	alarm(240); /* In case of emergency, please panic. */
 	while (1) {
 		if (fgets(buf, sizeof(buf), fp)==NULL) break;
 		while (buf[strlen(buf)-1]<' ' && buf[0]!='\0') buf[strlen(buf)-1]='\0';
@@ -191,7 +214,7 @@ void *tcpcgi_getdata(char *http_host, int http_port, char *http_path, char *cmd,
 				PRINTERR("Now expecting %i bytes of data.", datalen_loc);
 				retbuf=malloc(datalen_loc);
 				datalen_loc=fread(retbuf, 1, datalen_loc, fp);
-				PRINTERR("[%i]%s", datalen_loc, retbuf);
+				PRINTERR("[%i](omitted)", datalen_loc);
 			}
 			if (datalen) *datalen=datalen_loc;
 			PRINTERR("We got all that we want, let's get outta here.");
@@ -199,6 +222,8 @@ void *tcpcgi_getdata(char *http_host, int http_port, char *http_path, char *cmd,
 		}
 		PRINTERR("buf=%s", buf);
 	}
+	PRINTERR("Unscheduling panic alarm.");
+	alarm(0);
 
 	fclose(fp);
 	if (hex_cmd) free(hex_cmd);
@@ -221,6 +246,7 @@ char *tcpcgi_gensess(void) {
 		if (alphabet[j]=='\0' && i==0) { i--; continue; } /* Don't generate an empty session. */
 		ret[i]=alphabet[j];
 	}
+	ret[sizeof(ret)-1]='\0';
 	return(ret);
 }
 
@@ -228,7 +254,7 @@ void tcpcgi_closesession(int fd, char *http_host, int http_port, char *http_path
 	if (!session || fd<0) return;
 	PRINTERR("Terminating session \"%s\".", session);
 
-	tcpcgi_getdata(http_host, http_port, http_path, "close", session, NULL, NULL, NULL, NULL);
+	tcpcgi_getdata(http_host, http_port, http_path, "close", session, NULL, NULL, NULL, NULL, 0);
 	return;
 }
 
@@ -238,7 +264,7 @@ void tcpcgi_closesession(int fd, char *http_host, int http_port, char *http_path
  *    0   (no error, but no data was recvd)
  *    1   (no error, data recv)
  */
-int tcpcgi_handledata(char *http_host, int http_port, char *http_path, char *session, int fd, void *buf, int datalen) {
+int tcpcgi_handledata(char *http_host, int http_port, char *http_path, char *session, int proto_type, char *dest, int fd, void *buf, int datalen) {
 	unsigned char *ret;
 	unsigned char *lbuf=buf;
 	char *cmd;
@@ -247,7 +273,7 @@ int tcpcgi_handledata(char *http_host, int http_port, char *http_path, char *ses
 	int rv;
 
 	while (datalen>1024) {
-		if ((rv=tcpcgi_handledata(http_host, http_port, http_path, session, fd, lbuf, 1024))<0) {
+		if ((rv=tcpcgi_handledata(http_host, http_port, http_path, session, proto_type, dest, fd, lbuf, 1024))<0) {
 			return(rv);
 		}
 		datalen-=1024;
@@ -262,8 +288,8 @@ int tcpcgi_handledata(char *http_host, int http_port, char *http_path, char *ses
 		dataretlen=0;
 	}
 
-	ret=tcpcgi_getdata(http_host, http_port, http_path, cmd, session, NULL, lbuf, &dataretlen, &status);
-	PRINTERR("[%i]%s", dataretlen, ret);
+	ret=tcpcgi_getdata(http_host, http_port, http_path, cmd, session, dest, lbuf, &dataretlen, &status, proto_type);
+	PRINTERR("[%i](omitted, again)", dataretlen);
 	if (dataretlen>0 && ret && status>=0) {
 		write(fd, ret, dataretlen);
 		free(ret);
@@ -290,7 +316,7 @@ int handle_background(void) {
 		http_path=socks_info[i].http_path;
 		session=socks_info[i].session;
 		fd=socks_info[i].fd;
-		if ((rv=tcpcgi_handledata(http_host, http_port, http_path, session, fd, NULL, 0))<0) {
+		if ((rv=tcpcgi_handledata(http_host, http_port, http_path, session, SOCK_STREAM, NULL, fd, NULL, 0))<0) {
 			PRINTERR("[fd=%i] We got an error in the background and we don't know how to handle it!", fd);
 			close(fd);
 			continue;
@@ -302,9 +328,10 @@ int handle_background(void) {
 
 void sighandler(int sig) {
 	if (sig==SIGALRM) {
-		if (alarm_freq) alarm(alarm_freq);
-		handle_background();
+		PRINTERR("Recieved SIGALRM!");
+		return;
 	}
+	PRINTERR("Recieved weird signal %i", sig);
 }
 
 int print_help(char *msg) {
@@ -350,23 +377,28 @@ int print_help(char *msg) {
 #undef CLOSE_POLL
 #endif
 #define CLOSE_POLL(idx) { \
-	tcpcgi_closesession(socks_poll[idx].fd, http_host, http_port, http_path, socks_info[idx].session); \
-	close(socks_poll[idx].fd); \
-	socks_poll[idx].fd=-1; \
-	socks_info[idx].fd=-1; \
-	pollcnt--; \
-	pollpos=idx; \
+	if (idx!=0) { \
+		tcpcgi_closesession(socks_poll[idx].fd, http_host, http_port, http_path, socks_info[idx].session); \
+		close(socks_poll[idx].fd); \
+		socks_poll[idx].fd=-1; \
+		socks_info[idx].fd=-1; \
+		pollcnt--; \
+		pollpos=idx; \
+	} \
 }
 int main(int argc, char **argv) {
 	struct pollfd socks_poll[sizeof(socks_info)/sizeof(struct sockinfo)];
-	struct sockaddr addr;
+	struct sockaddr_in *addr;
+	struct sigaction sac;
 	signed int pollpos=0;
 	signed int status=0;
 	signed int ch;
 	char buf[8192];
 	char *http_host=NULL, *http_path=NULL;
 	char *hostport=NULL;
+	char udp_sess_wrk[128], *udp_sess, *socks_sess;
 	socklen_t addrlen;
+	sigset_t s;
 	int masterfd, fd, clientfd;
 	int pollcnt=0, pollmax;
 	int num_events;
@@ -462,8 +494,13 @@ int main(int argc, char **argv) {
 		return(print_help("Polltime max cannot be less than polltime min."));
 	}
 
+	sigemptyset(&s);
+	sigaddset(&s, SIGALRM);
+	sac.sa_handler=sighandler;
+	sac.sa_mask=s;
+	sac.sa_flags=0;
+	sigaction(SIGALRM, &sac, NULL);
 
-	signal(SIGALRM, sighandler);
 	signal(SIGPIPE, SIG_IGN);
 	masterfd=createlisten(listen_port, local_only, proto_type);
 	if (masterfd<0) {
@@ -497,7 +534,8 @@ int main(int argc, char **argv) {
 	pollpos++;
 	pollcnt++;
 
-	if (alarm_freq) alarm(1); /* Start scheduling background events. */
+	addrlen=sizeof(struct sockaddr_in);
+	addr=malloc(addrlen);
 
 	while (1) {
 		for (i=0; i<(sizeof(socks_poll)/sizeof(struct pollfd)); i++) {
@@ -548,11 +586,31 @@ int main(int argc, char **argv) {
 			fd=socks_poll[i].fd;
 			socks_poll[i].events=POLLIN;
 			socks_poll[i].revents=0;
-			if (fd==masterfd) {
+			/*
+			   This is special for UDP, since 
+			   it will never have a connection
+			   we have just guess at this by
+			   the existance (or lack of...)
+			   a session with a generated name
+			*/
+			if (proto_type==SOCK_DGRAM) {
+				PRINTERR("Incoming UDP ...");
+				if (fd!=masterfd) {
+					PRINTERR("... not from master?! IGNORING!");
+					continue;
+				}
+				addrlen=sizeof(struct sockaddr_in);
+				recvfrom(fd, buf, 5, MSG_PEEK, (struct sockaddr *)addr, &addrlen);
+				memcpy(udp_sess_wrk, &addr->sin_port, 2);
+				memcpy(udp_sess_wrk+2, &addr->sin_addr.s_addr, sizeof(addr->sin_addr.s_addr));
+				udp_sess=hexcode(udp_sess_wrk, 2+sizeof(addr->sin_addr.s_addr));
+				PRINTERR("Session name is .. %s", udp_sess);
+			}
+			if (fd==masterfd && proto_type!=SOCK_DGRAM) {
 				/* Handle incoming connection. */
 				PRINTERR("New connection...");
-				addrlen=sizeof(addr);
-				clientfd=accept(fd, &addr, &addrlen);
+				addrlen=sizeof(struct sockaddr_in);
+				clientfd=accept(fd, (struct sockaddr *)addr, &addrlen);
 				if (clientfd<0) {
 					PERROR("accept");
 					continue;
@@ -581,7 +639,7 @@ int main(int argc, char **argv) {
 				socks_info[pollpos].http_port=http_port;
 				socks_info[pollpos].http_host=http_host;
 				socks_info[pollpos].http_path=http_path;
-				tcpcgi_getdata(http_host, http_port, http_path, "open", socks_info[pollpos].session, hostport, NULL, NULL, &status);
+				tcpcgi_getdata(http_host, http_port, http_path, "open", socks_info[pollpos].session, hostport, NULL, NULL, &status, proto_type);
 				if (status<0) {
 					PRINTERR("Error creating session on remote end.");
 					CLOSE_POLL(pollpos);
@@ -594,36 +652,45 @@ int main(int argc, char **argv) {
 
 				PRINTERR("... from [idx=%i, fd=%i], pollcnt=%i, pollmax=%i: session=%s", pollpos, clientfd, pollcnt, pollmax, socks_info[pollpos].session);
 				pollpos=-1;
+				if (num_events==0) break;
+				continue;
+			}
+			/* Handle data from an existing connection. */
+			PRINTERR("Data waiting on [idx=%i, fd=%i]...", i, fd);
+			x=read(fd, buf, sizeof(buf));
+			PRINTERR(".. not anymore, we read %i bytes.", x);
+			if (x<0) PERROR("read");
+			if (x==0) {
+				PRINTERR("EOF from [idx=%i, fd=%i].", i, fd);
+				CLOSE_POLL(i);
+				continue;
+			}
+			if (proto_type==SOCK_DGRAM) {
+				socks_sess=udp_sess;
 			} else {
-				/* Handle data from an existing connection. */
-				PRINTERR("Data waiting on [idx=%i, fd=%i]...", i, fd);
-				x=read(fd, buf, sizeof(buf));
-				if (x<0) PERROR("read");
-				if (x==0) {
-					PRINTERR("EOF from [idx=%i, fd=%i].", i, fd);
-					CLOSE_POLL(i);
-					continue;
-				}
-				if ((rv=tcpcgi_handledata(http_host, http_port, http_path, socks_info[i].session, fd, buf, x))<0) {
-					PRINTERR("Unable to put data.");
-					CLOSE_POLL(i);
-					continue;
-				}
-				if (poll_backoff_ireset) {
-					/* Check background activity, and adjust polltime if needed. */
-					if (handle_background()==0) {
-						/* If nothing is happening in the background, slow down. */
-						if (poll_backoff<5) { poll_time*=poll_backoff; } else { poll_time+=poll_backoff; }
-						PRINTERR("Lengthening polltime [poll_time=%i], nothing is happening.", poll_time);
-					} else {
-						/* If we're active in the background, speed up! */
-						if (poll_backoff<5) { poll_time/=poll_backoff; } else { poll_time-=poll_backoff; }
-						if (poll_backoff_reset) poll_time=poll_time_min;
-						PRINTERR("Shortening polltime [poll_time=%i], stuff is happening.", poll_time);
-					}
-					if (poll_time>poll_time_max) poll_time=poll_time_max;
-					if (poll_time<poll_time_min) poll_time=poll_time_min;
+				socks_sess=socks_info[i].session;
+			}
+
+			/* Do the actual transmission */
+			if ((rv=tcpcgi_handledata(http_host, http_port, http_path, socks_sess, proto_type, hostport, fd, buf, x))<0) {
+				PRINTERR("Unable to put data.");
+				CLOSE_POLL(i);
+				continue;
+			}
+			if (poll_backoff_ireset) {
+				/* Check background activity, and adjust polltime if needed. */
+				if (handle_background()==0) {
+					/* If nothing is happening in the background, slow down. */
+					if (poll_backoff<5) { poll_time*=poll_backoff; } else { poll_time+=poll_backoff; }
+					PRINTERR("Lengthening polltime [poll_time=%i], nothing is happening.", poll_time);
+				} else {
+					/* If we're active in the background, speed up! */
+					if (poll_backoff<5) { poll_time/=poll_backoff; } else { poll_time-=poll_backoff; }
+					if (poll_backoff_reset) poll_time=poll_time_min;
+					PRINTERR("Shortening polltime [poll_time=%i], stuff is happening.", poll_time);
 				}
+				if (poll_time>poll_time_max) poll_time=poll_time_max;
+				if (poll_time<poll_time_min) poll_time=poll_time_min;
 			}
 			if (num_events==0) break;
 		}
