@@ -19,14 +19,14 @@
 	email: tcpcgi@rkeene.org
 
 */
-
+#include <string.h>
+#define _XOPEN_SOURCE_EXTENDED
 #include <sys/socket.h>
 #include <sys/types.h>
 #include <sys/time.h>
 #include <sys/poll.h>
 #include <sys/wait.h>
 #include <unistd.h>
-#include <string.h>
 #include <stdlib.h>
 #include <signal.h>
 #include <stdio.h>
@@ -116,15 +116,12 @@ int tcpcgi_cmd_open(int fd, cache_t *parms) {
 	struct sessioninfo *si;
 	char *host, *hostport, *port_str;
 	char *session;
-	char *proto;
-	int proto_type;
 	int port;
 	int sockfd;
 
 	PRINTERR("Starting tcpcgi_cmd_open");
 	hostport=cache_find(parms, "dest", strlen("dest"), NULL);
 	session=cache_find(parms, "session", strlen("session"), NULL);
-	proto=cache_find(parms, "proto", strlen("proto"), NULL);
 	if (!hostport) return(tcpcgi_errmsg(fd, "Need parameter: dest"));
 	if (!session) return(tcpcgi_errmsg(fd, "Need parameter: session"));
 	si=cache_find(sess, session, strlen(session), NULL);
@@ -140,14 +137,7 @@ int tcpcgi_cmd_open(int fd, cache_t *parms) {
 
 	/* Connect to host:port */
 	PRINTERR("[%s] Connecting to \"%s\", port %i", session, host, port);
-	if (proto) {
-		proto_type=SOCK_STREAM;
-		if (strcmp(proto,"tcp")==0) proto_type=SOCK_STREAM;
-		if (strcmp(proto,"udp")==0) proto_type=SOCK_DGRAM;
-		sockfd=createconnection(host, port, proto_type);
-	} else {
-		sockfd=createconnection_tcp(host, port);
-	}
+	sockfd=createconnection_tcp(host, port);
 	if (sockfd<0) return(tcpcgi_errmsg(fd, "Unable to connect"));
 
 	/* Let 'em know we did something. */
@@ -241,23 +231,41 @@ int tcpcgi_cmd_get(int fd, cache_t *parms) {
 
 int tcpcgi_cmd_put(int fd, cache_t *parms) {
 	struct sessioninfo *si;
-	char *session, *data;
+	char *session, *data, *proto, *dest, *dest_addr;
+	int dest_port;
+	int proto_type;
 	int datalen=0;
 	int x;
 
 	PRINTERR("Starting tcpcgi_cmd_put");
+	proto=cache_find(parms, "proto", strlen("proto"), NULL);
 	session=cache_find(parms, "session", strlen("session"), NULL);
 	data=cache_find(parms, "data", strlen("data"), &datalen);
 	if (!session) return(tcpcgi_errmsg(fd, "Need parameter: session"));
 	if (!data) return(tcpcgi_errmsg(fd, "Need parameter: data"));
+	if (!proto) proto="tcp";
+	proto_type=SOCK_STREAM;
+	if (strcmp(proto,"udp")==0) proto_type=SOCK_DGRAM;
 
-	si=cache_find(sess, session, strlen(session), NULL);
-	if (!si) return(tcpcgi_errmsg(fd, "Improper parameter: session;  Session does not exists"));
-
-	while ((x=write(si->sockfd, data, datalen))>0) {
-		datalen-=x;
-		if (datalen==0) break;
+	if (proto_type!=SOCK_DGRAM) {
+		si=cache_find(sess, session, strlen(session), NULL);
+		if (!si) return(tcpcgi_errmsg(fd, "Improper parameter: session;  Session does not exists"));
+		while ((x=write(si->sockfd, data, datalen))>0) {
+			datalen-=x;
+			if (datalen==0) break;
+		}
+	} else {
+		dest=cache_find(parms, "dest", strlen("dest"), NULL);
+		if (!dest) return(tcpcgi_errmsg(fd, "Need parameter: dest"));
+		if (strchr(dest,':')==NULL) return(tcpcgi_errmsg(fd, "Improper parameter: dest"));
+		dest_addr=strdup(dest);
+		dest_port=atoi(strchr(dest,':')+1);
+		*(strchr(dest_addr,':'))='\0';
+		PRINTERR("dest=%s, dest_port=%i, dest_addr=%s", dest, dest_port, dest_addr);
+		/* XXX:  WORK ON UDP SUPPORT WAS TEMPORARILY SUSPENDED HERE ON 4Sep03 */
 	}
+
+
 	if (x<0) {
 		/* We destroy the session at this point ... */
 		cache_delete(sess, session, strlen(session), NULL, 1);
